#include <gtk/gtk.h>
#include <gtk/gtkoptionmenu.h>
#include <string.h>        /* memcpy */
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include "colorconv.h"

#include "ui.h"
#include "color_correct.h"
#include "data.h"

typedef struct
  {
  gint run;
  }
yuvInterface;

gchar *PLUG_IN_NAME[] =
{
  "plug_in_color_correct",
  "plug_in_autowb"
};

#define CC_CC    0
#define CC_AUTOWB  1

gint cc_type = CC_CC;

static gint32 volatile  image_ID_global     = -1;
GimpDrawable           *src_drawable_global = NULL;

/* Declare local functions.
 */
static void query (void);
static void run (const gchar       *name,
         gint               nparams,
         const GimpParam   *param,
         gint              *nreturn_vals,
                 GimpParam        **return_vals);

static void yuvadjust (GimpDrawable * drawable);

static gint yuvadjust_dialog (GimpDrawable * drawable);

static void prepare_correction (void);

void update_image_preview (void);
void destroy_image_preview (void);

/***** Local vars *****/

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,            /* init_proc  */
  NULL,            /* quit_proc  */
  query,            /* query_proc */
  run,            /* run_proc   */
};


MAIN ()

static void
query (void)
{
  static GimpParamDef args[] = {
    {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
    {GIMP_PDB_IMAGE, "image", "Input image"},
    {GIMP_PDB_DRAWABLE, "drawable", "Input drawable"},

    {GIMP_PDB_INT32, "black_u",
     "blackpoint U-component-adjust (-128 .. 127 - default 0)"},
    {GIMP_PDB_INT32, "black_v",
     "blackpoint V-component-adjust (-128 .. 127 - default 0)"},

    {GIMP_PDB_INT32, "gray_u",
     "graypoint U-component-adjust (-128 .. 127 - default 0)"},
    {GIMP_PDB_INT32, "gray_v",
     "graypoint V-component-adjust (-128 .. 127 - default 0)"},

    {GIMP_PDB_INT32, "white_u",
     "whitepoint U-component-adjust (-128 .. 127 - default 0)"},
    {GIMP_PDB_INT32, "white_v",
     "whitepoint V-component-adjust (-128 .. 127 - default 0)"},

    {GIMP_PDB_INT32, "saturation",
     "saturation (0 .. 400) - default 100 (=1.00) )"},

    {GIMP_PDB_INT32, "black_in",
     "black input value in levels (0..255 - default 0)"},
    {GIMP_PDB_INT32, "gray_in",
     "gray input value in levels (0..255 - default 128)"},
    {GIMP_PDB_INT32, "white_in",
     "white input value in levels (0..255 - default 255)"},

    {GIMP_PDB_INT32, "black_out",
     "black output value in levels (0..255 - default 0)"},
    {GIMP_PDB_INT32, "white_out",
     "white output value in levels (0..255 - default 255)"}
  };

  const gchar *author = "OEyvind Kolaas <pippin@users.sf.net>";
  const gchar *copyrights = "OEyvind Kolaas";
  const gchar *copyright_date = "2002";

  gimp_install_procedure (PLUG_IN_NAME[CC_CC],
              "Adjust the colors of the image in yuv colorspace",
              "Adjusts the colors of the image in yuv colorspace.",
              (gchar *) author,
              (gchar *) copyrights, (gchar *) copyright_date,

              N_("<Image>/Layer/Colors/Color Correction..."),
              "RGB*",
              GIMP_PLUGIN, sizeof (args) / sizeof (args[0])
              , 0, args, NULL);
}


static void
run (const gchar      *name,
     gint              nparams,
     const GimpParam  *param,
     gint             *nreturn_vals,
     GimpParam       **return_vals)
{
  static GimpParam values[1];
  GimpDrawable *drawable;

#ifdef GIMP_1_2
  GimpRunModeType run_mode;
#else
  GimpRunMode run_mode;
#endif
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;

  run_mode = param[0].data.d_int32;


  if (!strcmp (name, PLUG_IN_NAME[CC_CC]))
    cc_type = CC_CC;
  else if (!strcmp (name, PLUG_IN_NAME[CC_AUTOWB]))
    cc_type = CC_AUTOWB;

  /*  Get the specified drawable  */

  image_ID_global = param[1].data.d_int32;
  src_drawable_global = drawable =
    gimp_drawable_get (param[2].data.d_drawable);

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;


  switch (run_mode) {
    case GIMP_RUN_INTERACTIVE:
      /*  Possibly retrieve data  */
      gimp_get_data (PLUG_IN_NAME[cc_type], &pp);

      switch (cc_type) {
        case CC_CC:
          gimp_image_undo_freeze (image_ID_global);

          if (!yuvadjust_dialog (drawable)) {
            gimp_image_undo_thaw (image_ID_global);
            return;
          }

          gimp_image_undo_thaw (image_ID_global);
          break;
        case CC_AUTOWB:
          calculate_greyworld ();
          break;
        default:
          break;
      }
      break;

    case GIMP_RUN_NONINTERACTIVE:
      /*  Make sure all the arguments are there!  */
      switch (cc_type) {
        case CC_CC:
          if (nparams != 15) {
            status = GIMP_PDB_CALLING_ERROR;
          } else {
            pv.bu = param[3].data.d_int32;
            pv.bv = param[4].data.d_int32;
            pv.gu = param[5].data.d_int32;
            pv.gv = param[6].data.d_int32;
            pv.wu = param[7].data.d_int32;
            pv.wv = param[8].data.d_int32;
            pv.saturation = param[9].data.d_int32;
            pv.by = param[10].data.d_int32;
            pv.gy = param[11].data.d_int32;
            pv.wy = param[12].data.d_int32;
            pv.outb = param[13].data.d_int32;
            pv.outw = param[14].data.d_int32;
          }

          /*FIXME: add checks on values */
          if ((status == GIMP_PDB_SUCCESS) && pv.outb > 255)

            status = GIMP_PDB_CALLING_ERROR;
          break;
        case CC_AUTOWB:
          if (nparams != 3) ;  /*what me worry? */
          break;
        default:
          break;
      }


      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Possibly retrieve data  */
      gimp_get_data (PLUG_IN_NAME[cc_type], &pp);
      break;

    default:
      break;
  }

  if (status == GIMP_PDB_SUCCESS) {
    /*  Make sure that the drawable is RGB color  */

    if (gimp_drawable_is_rgb (drawable->drawable_id)) {

      gimp_image_undo_group_start (image_ID_global);
      yuvadjust (drawable);
      gimp_image_undo_group_end (image_ID_global);

      if (run_mode != GIMP_RUN_NONINTERACTIVE)
        gimp_displays_flush ();

      /*  Store data  */
      if (run_mode == GIMP_RUN_INTERACTIVE)
        gimp_set_data (PLUG_IN_NAME[cc_type], &pp,
                 sizeof (persistantValues));
    } else {
      /* gimp_message ("color_corrector: only operates on RGB images"); */
      /* (why,.. it might as well operate on the LUT of palette images,..*/
      status = GIMP_PDB_EXECUTION_ERROR;
    }
  }

  values[0].data.d_status = status;

  gimp_drawable_detach (drawable);
}


static gint
yuvadjust_dialog (GimpDrawable * drawable)
{
    gboolean run;
  GtkWidget *dialog;

  gimp_ui_init ("yuv_adjust", FALSE);

  dialog = color_correct_dialog ();
  gtk_widget_show (dialog);

    switch (run = gimp_dialog_run (GIMP_DIALOG (dialog))) {
        case GTK_RESPONSE_OK:
            break;
        case GTK_RESPONSE_CANCEL:
            break;
    }

    gtk_widget_destroy (dialog);
    return run;
}

void
yuvadjust_close_callback (GtkWidget *widget,
                          gpointer   data)
{
  destroy_image_preview ();
  gimp_displays_flush ();
  gtk_main_quit ();
}

void calculate_greyworld (void)
{
  GimpPixelRgn rgn;
  gpointer pr;
  gint Ysum = 0, Usum = 0, Vsum = 0;
  gint count = 0;

  gimp_pixel_rgn_init (&rgn, src_drawable_global, 0, 0,
             src_drawable_global->width,
             src_drawable_global->height, FALSE, FALSE);

  for (pr = gimp_pixel_rgns_register (1, &rgn);
     pr != NULL; pr = gimp_pixel_rgns_process (pr)) {
    int x, y;
    guchar *src = rgn.data;

    for (y = 0; y < rgn.h; y++) {
      guchar *s = src;

      for (x = 0; x < rgn.w; x++) {
        int R, G, B, Y, U, V;

        R = s[0];
        G = s[1];
        B = s[2];
        RGB2YUV (R, G, B, Y, U, V);
        Ysum += Y;
        Usum += U;
        Vsum += V;
        count++;

        s += rgn.bpp;
      }
      src += rgn.rowstride;
    }
  }

  Ysum /= count;
  Usum /= count;
  Vsum /= count;

  pv.bu = pv.bv = 0;
  pv.wu = -(Usum - 128)*2;
  pv.wv = -(Vsum - 128)*2;
  pv.gu = (pv.bu + pv.wu) / 2;
  pv.gv = (pv.bv + pv.wv) / 2;
}

static int RGB_Lut[256];
static int U_Lut[256];
static int V_Lut[256];

static void prepare_correction (void)
{
  int x;

  for (x = 0; x < 256; x++) {
    int in_scale = clamp (x < pv.by ? 0 :
                x > pv.wy ? 255 :
                x == pv.gy ? 128 :
                x <
                pv.gy ? (x - pv.by) * 128 / (pv.gy -
                               pv.by) : (x -
                                   pv.gy) *
                128 / (pv.wy - pv.gy) + 128, 0, 255);
    RGB_Lut[x] =
      clamp (pv.outb + in_scale * (pv.outw - pv.outb) / 256, 0, 255);
    U_Lut[x] =
      (x <
       128 ? ((x * (pv.gu - pv.bu)) / 128) +
       pv.bu : (((x - 128) * (pv.wu - pv.gu)) / 128) + pv.gu);
    V_Lut[x] =
      (x <
       128 ? ((x * (pv.gv - pv.bv)) / 128) +
       pv.bv : (((x - 128) * (pv.wv - pv.gv)) / 128) + pv.gv);
  }
}

static void process_tile (GimpPixelRgn *srgn,
                          GimpPixelRgn *drgn,
                          gboolean      vectorscope)
{
  gint x, y;
  guchar *src, *s;
  guchar *dest, *d;

  src = srgn->data;
  dest = drgn->data;

  for (y = 0; y < srgn->h; y++) {
    s = src;
    d = dest;

    for (x = 0; x < srgn->w; x++) {
      int R, G, B;

      R = s[0];
      G = s[1];
      B = s[2];
      R = RGB_Lut[R];
      G = RGB_Lut[G];
      B = RGB_Lut[B];

      {
        int Y, U, V;

        RGB2YUV (R, G, B, Y, U, V);
        U += U_Lut[Y];
        V += V_Lut[V];
        U = (((U - 128) * pv.saturation) / 100) + 128;
        V = (((V - 128) * pv.saturation) / 100) + 128;

        YUV2RGB (Y, U, V, R, G, B);

        if (vectorscope) {
          int x, y;
          gchar *v;

          x = (((U - 128) / 2) + 64);
          y = (127 - (((V - 128) / 2) + 64));
          v = &vectorscope_data[3 *
                      (128 * clamp (y, 0, 127) +
                       clamp (x, 0, 127))];
          v[0] = R;
          v[1] = G;
          v[2] = B;
        }
      }
      d[0] = R;
      d[1] = G;
      d[2] = B;
      if (srgn->bpp >= 4)
        d[3] = s[3];
      else
        d[3] = 255;
      s += srgn->bpp;
      d += drgn->bpp;
    }
    src += srgn->rowstride;
    dest += drgn->rowstride;
  }
}

static void yuvadjust    (GimpDrawable *drawable)
{
  gint x1, y1, x2, y2;
  gpointer pr;

  GimpPixelRgn srgn, drgn;

  prepare_correction ();

  gimp_drawable_mask_bounds (drawable->drawable_id, &x1, &y1, &x2, &y2);

  gimp_pixel_rgn_init (&srgn, drawable, 0, 0,
             drawable->width, drawable->height, FALSE, FALSE);
  gimp_pixel_rgn_init (&drgn, drawable, 0, 0,
             drawable->width, drawable->height, TRUE, TRUE);

  for (pr = gimp_pixel_rgns_register (2, &srgn, &drgn);
     pr != NULL; pr = gimp_pixel_rgns_process (pr)) {

    process_tile (&srgn, &drgn, FALSE);
  }

  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
  gimp_drawable_update (drawable->drawable_id, x1, y1, (x2 - x1),
              (y2 - y1));
}

/*
 *  utility function, which probably could be easier implemented internally in
 *  the gimp, or perhaps there is one and I just don't know about it,..
 */
static void clear_drawable (GimpDrawable * drawable)
{
  GimpPixelRgn rgn;
  gpointer pr;

  if (!drawable)
    return;

  gimp_pixel_rgn_init (&rgn, drawable, 0, 0,
             drawable->width, drawable->height, TRUE, FALSE);

  for (pr = gimp_pixel_rgns_register (1, &rgn);
     pr != NULL; pr = gimp_pixel_rgns_process (pr))
    memset (rgn.data, 0, rgn.rowstride * rgn.h);
}

/*****************************************************************************************
 **
 **  All image area background preview code is below this commentblock
 **
 *****************************************************************************************
 *****************************************************************************************/

#define INIT    0
#define FINISH  1
#define PROCESS  2
#define ABORT    3
#define DEAD    4

typedef struct {
  gint state;          /* the state our task is in */

  gint x1;
  gint x2;
  gint y1;
  gint y2;
  guchar *buffer;
  guchar *ibuffer;
  GimpPixelRgn srgn;
  GimpPixelRgn drgn;
  GimpDrawable *drawable;
  gint32 layer_ID;
  gpointer pr;

} BackgroundPersistent;

static gint *task_state = NULL;

static gint32 preview_layer_ID = -1;
static GimpDrawable *preview_drawable = NULL;

static gboolean background_preview (BackgroundPersistent * bp);


/*  Update the state of the image-window to reflect plugin values.
 *
 */
void update_image_preview (void)
{
  if (!pp.prev_in_image) {
    destroy_image_preview ();
  } else {
    if (preview_layer_ID == -1) {  /* create the preview layer */
      preview_layer_ID =
        gimp_layer_new (image_ID_global, _("preview layer"),
                src_drawable_global->width,
                src_drawable_global->height, GIMP_RGBA_IMAGE,
                100, GIMP_NORMAL_MODE);
      preview_drawable = gimp_drawable_get (preview_layer_ID);

      {          /* must manually clear the tiles of the drawable
                   if there is a selection */
        gboolean is_selection;
        gint dummy;

        gimp_selection_bounds (image_ID_global,
                     &is_selection, &dummy, &dummy, &dummy,
                     &dummy);
        if (is_selection)
          clear_drawable (preview_drawable);
      }

      gimp_image_add_layer (image_ID_global, preview_layer_ID, 0);
    }

    if (task_state)      /* signal the background task to stop */
      *task_state = ABORT;

    {            /* start new background job */
      BackgroundPersistent *bp = g_new (BackgroundPersistent, 1);

      task_state = &(bp->state);
      bp->state = INIT;
      bp->drawable = preview_drawable;

      gtk_idle_add ((GtkFunction) background_preview, (gpointer) bp);
    }
  }
}

/*
 *  cleans up whatever nastyness multiple calls to update_image_preview has donw
 */
void destroy_image_preview (void)
{
  if (preview_layer_ID != -1) {
    gimp_drawable_detach (preview_drawable);
    gimp_image_remove_layer (image_ID_global, preview_layer_ID);
    preview_layer_ID = -1;

    gimp_displays_flush ();  /* make gimp reflect our changes */
  }

  if (task_state)
    *task_state = ABORT;  /* signal the background task to stop */
}

/*
 *  The background "thread" that is responsible for creating the preview layer
 *
 */
gboolean background_preview (BackgroundPersistent * bp)
{
  switch (bp->state) {
    case INIT:
      prepare_correction ();  /* make lookuptables */

      gimp_drawable_mask_bounds (src_drawable_global->drawable_id,
                     &bp->x1, &bp->y1, &bp->x2, &bp->y2);

      {
        gint width = bp->x2 - bp->x1;
        gint height = bp->y2 - bp->y1;

        gimp_pixel_rgn_init (&bp->srgn, src_drawable_global, bp->x1,
                   bp->y1, width, height, FALSE, FALSE);

        gimp_pixel_rgn_init (&bp->drgn, bp->drawable, bp->x1, bp->y1,
                   width, height, TRUE, FALSE);
      }

      bp->pr = gimp_pixel_rgns_register (2, &bp->srgn, &bp->drgn);

      if (pp.vectorscope_enable) {
        memset (vectorscope_data, 0, 128 * 128 * 3);
      }

      bp->state = PROCESS;
      return TRUE;
    case PROCESS:
      while (bp->pr) {
        process_tile (&bp->srgn, &bp->drgn, pp.vectorscope_enable);

        bp->pr = gimp_pixel_rgns_process (bp->pr);

        return TRUE;
      }
      bp->state = FINISH;
    case FINISH:
      gimp_drawable_flush (bp->drawable);

      gimp_drawable_update (bp->drawable->drawable_id,
                  bp->x1, bp->y1, (bp->x2 - bp->x1),
                  (bp->y2 - bp->y1));

      gtk_widget_queue_draw (vectorscope_drawingarea);

      gimp_displays_flush ();

    case ABORT:
      if (task_state == &(bp->state))
        task_state = NULL;

      g_free (bp);

      return FALSE;
    case DEAD:
    default:
      g_message ("call 911");
      return FALSE;
  }
}
